@@ -11,10 +11,13 @@ use POSIX qw//;
 use Encode qw//;
 use XML::LibXML;
 use Interchange::Search::Solr::Response;
+use Interchange::Search::Solr::Builder;
 use Lingua::StopWords;
 use Types::Standard qw/ArrayRef HashRef Int Bool/;
 use namespace::clean;
 use HTTP::Response;
+use Scalar::Util;
+use constant { DEBUG => 0 };
 
 =head1 NAME
 
@@ -22,11 +25,11 @@ Interchange::Search::Solr -- Solr query encapsulation
 
 =head1 VERSION
 
-Version 0.12
+Version 0.13
 
 =cut
 
-our $VERSION = '0.12';
+our $VERSION = '0.13';
 
 =head1 DESCRIPTION
 
@@ -162,9 +165,17 @@ L<Webservice::Solr::Query> for stringification.
 The field used to sort the result (optional and defaults to score, as
 per Solr doc).
 
+You can set it to a scalar with a field name or instead you can use
+the L<SQL::Abstract> syntax (all the cases documented there are
+supported). E.g.
+
+ $solr->sorting([{ -asc => 'created_date' }, {-desc => [qw/updated_date sku/] }]);
+
+If you pass a reference, the C<sorting_direction> setting is ignored.
+
 =head2 sorting_direction
 
-The direction used by the sorting, when C<sorting> is specified.
+The direction used by the sorting, when C<sorting> is specified and is a plain scalar.
 Default to 'desc'.
 
 =cut
@@ -335,6 +346,22 @@ sub _build_solr_object {
     return WebService::Solr->new(@args);
 }
 
+=head2 builder_object(\@terms, \%filters, $page)
+
+Creates Interchange::Search::Solr::Builder instance.
+
+=cut
+
+sub builder_object {
+    my ($self, $terms, $filters, $page) = @_;
+    return Interchange::Search::Solr::Builder->new(
+        terms   => $terms,
+        filters => $filters,
+        facets  => $self->facets,
+        page    => $page
+    );
+}
+
 =head1 METHODS
 
 =head2 search( [ $string_or$structure ] ])
@@ -476,8 +503,13 @@ sub execute_query {
     }
     unless ($our_res) {
         my %params = $self->construct_params;
+        # print Dumper(\%params) if DEBUG;
         my $res = $self->solr_object->search($querystring, \%params);
         $our_res = Interchange::Search::Solr::Response->new($res->raw_response);
+
+	if ($our_res->solr_status != 0) {
+	    die "Solr failure: ".$our_res->raw_response->message;
+	}
     }
     $self->_set_response($our_res);
     $self->permit_empty_search(0);
@@ -521,7 +553,14 @@ sub construct_params {
         }
     }
     if (my $sort_by = $self->sorting) {
-        $params{sort} = join(' ', $sort_by, $self->sorting_direction);
+        my $sort_by_struct;
+        if (ref($sort_by)) {
+            $sort_by_struct = $sort_by;
+        }
+        else {
+            $sort_by_struct = { '-' . $self->sorting_direction => $sort_by };
+        }
+        $params{sort} = join(', ', $self->_build_sort_field($sort_by_struct));
     }
     if (my $fl = $self->return_fields) {
         $params{fl} = join(',', @$fl);
@@ -739,8 +778,11 @@ sub add_terms_to_url {
     my @terms = @{ $self->search_terms };
     push @terms, @additional_terms;
     $self->search_terms(\@terms);
-    return $self->url_builder($self->search_terms,
-                              $self->filters);
+    my $builder =  $self->builder_object(
+        $self->search_terms,
+        $self->filters
+    );
+    return $builder->url_builder;
     
 }
 
@@ -840,35 +882,11 @@ sub current_search_to_url {
         $page = $self->page;
     }
 
-    return $self->url_builder($self->search_terms,
+    my $builder = $self->builder_object($self->search_terms,
                               $self->filters,
                               $page);
-}
 
-=head2 url_builder(\@terms, \%filters, $page);
-
-Build a query url with the parameter passed
-
-=cut
-
-
-sub url_builder {
-    my ($self, $terms, $filters, $page) = @_;
-    my @fragments;
-    if (@$terms) {
-        push @fragments, 'words', @$terms;
-    }
-    if (%$filters) {
-        foreach my $facet (@{ $self->facets }) {
-            if (my $terms = $filters->{$facet}) {
-                push @fragments, $facet, @$terms;
-            }
-        }
-    }
-    if ($page and $page > 1) {
-        push @fragments, page => $page;
-    }
-    return join ('/', @fragments);
+    return $builder->url_builder;
 }
 
 sub _build_facet_url {
@@ -902,8 +920,8 @@ sub _build_facet_url {
         $toggled_filters{$facet} = \@active if @active;
     }
     #    print Dumper(\@terms, \%toggled_filters);
-    my $url = $self->url_builder(\@terms, \%toggled_filters);
-    return $url;
+    my $builder = $self->builder_object(\@terms, \%toggled_filters);
+    return $builder->url_builder;
 }
 
 sub _filter_is_active {
@@ -984,11 +1002,12 @@ sub paginator {
     my $end   = ($page + $page_scope < $total_pages) ? ($page + $page_scope) : $total_pages;
 
     my %pager = (items => []);
+    my $builder = $self->builder_object($self->search_terms, $self->filters);
+
     for (my $count = $start; $count <= $end ; $count++) {
         # create the link
-        my $url = $self->url_builder($self->search_terms,
-                                     $self->filters,
-                                     $count);
+        $builder->page($count);
+        my $url = $builder->url_builder;
         my $item = {
                     url => $url,
                     name => $count,
@@ -1008,14 +1027,13 @@ sub paginator {
         push @{$pager{items}}, $item;
     }
     if ($page != $total_pages) {
-        $pager{last} = $self->url_builder($self->search_terms,
-                                          $self->filters,
-                                          $total_pages);
+        $builder->page($total_pages);
+        $pager{last} = $builder->url_builder;
         $pager{last_page} = $total_pages;
     }
     if ($page != 1) {
-        $pager{first} = $self->url_builder($self->search_terms,
-                                           $self->filters, 1);
+        $builder->page(1);
+        $pager{first} = $builder->url_builder;
         $pager{first_page} = 1;
     }
     $pager{total_pages} = $total_pages;
@@ -1059,15 +1077,17 @@ sub terms_found {
     my @terms = @{ $self->search_terms };
     return unless @terms;
     my %out = (
-               reset => $self->url_builder([], $self->filters),
+               reset => $self->builder_object([], $self->filters)->url_builder,
                terms => [],
               );
+    my @toggled;
+    my $builder = $self->builder_object(\@toggled, $self->filters);
     foreach my $term (@terms) {
-        my @toggled = grep { $_ ne $term } @terms;
+        @toggled = grep { $_ ne $term } @terms;
+        $builder->terms(\@toggled);
         push @{ $out{terms} }, {
                                 term => $term,
-                                url => $self->url_builder(\@toggled,
-                                                          $self->filters),
+                                url  => $builder->url_builder,
                                };
     }
     return \%out;
@@ -1173,6 +1193,73 @@ sub _term_is_good {
     return 0;
 }
 
+# stolen from SQL::Abstract
+
+sub _build_sort_field {
+    my ($self, $arg) = @_;
+    return $self->_SWITCH_refkind($arg,
+                                  {
+                                   ARRAYREF => sub {
+                                       map { $self->_build_sort_field($_) } @$arg;
+                                   },
+                                   SCALAR => sub {
+                                       return "$arg";
+                                   },
+                                   UNDEF => sub {
+                                       return;
+                                   },
+                                   SCALARREF => sub {
+                                       $$arg;
+                                   },
+                                   HASHREF => sub {
+                                       my ($key, $val, @rest) = %$arg;
+                                       return () unless $key;
+                                       if ( @rest or not $key =~ /^-(desc|asc)/i ) {
+                                           die "hash passed to sorting  must have exactly ".
+                                             "one key (-desc or -asc)";
+                                       }
+                                       my $direction = $1;
+                                       my @ret;
+                                       for my $c ($self->_build_sort_field($val)) {
+                                           my $query;
+                                           $self->_SWITCH_refkind ($c,
+                                                                   {
+                                                                    SCALAR => sub {
+                                                                        $query = $c;
+                                                                    },
+                                                                   });
+                                           $query = $query . ' ' . lc($direction);
+                                           push @ret, $query;
+                                       }
+                                       return @ret;
+                                   },
+                                  });
+}
+
+
+sub _refkind {
+    my ($self, $data) = @_;
+    return 'UNDEF' unless defined $data;
+    # blessed objects are treated like scalars
+    my $ref = (Scalar::Util::blessed $data) ? '' : ref $data;
+    return 'SCALAR' unless $ref;
+    my $n_steps = 1;
+    while ($ref eq 'REF') {
+        $data = $$data;
+        $ref = (Scalar::Util::blessed $data) ? '' : ref $data;
+        $n_steps++ if $ref;
+    }
+    return ($ref||'SCALAR') . ('REF' x $n_steps);
+}
+
+sub _SWITCH_refkind {
+    my ($self, $data, $dispatch_table) = @_;
+    my $type = $self->_refkind($data);
+    my $coderef = $dispatch_table->{$self->_refkind($data)};
+    die "Unsupported structure $type" unless $coderef;
+    $coderef->();
+}
+
 
 =head1 AUTHOR
 
